/*******************************************************************************
 * Copyright 2010 Hon Hai Precision Ind. Co. Ltd.
 * All Rights Reserved.
 * No portions of this material shall be reproduced in any form without the
 * written permission of Hon Hai Precision Ind. Co. Ltd.
 *
 * All information contained in this document is Hon Hai Precision Ind.
 * Co. Ltd. company private, proprietary, and trade secret property and
 * are protected by international intellectual property laws and treaties.
 *
 ******************************************************************************/
 
 
#ifndef _BR_DNS_HIJACK_H_
#define _BR_DNS_HIJACK_H_

#ifndef uint16
#define uint16 unsigned short
#endif
#ifndef uint8
#define uint8 unsigned char
#endif
#ifndef uint32
#define uint32 unsigned int
#endif
#ifndef int32
#define int32 int
#endif
#ifndef int8
#define int8 char
#endif

#ifndef int16
#define int16 short
#endif

#define MS_ACCEPT       0
#define MS_DROP         1

#define BR_DROP     0
#define BR_ACCEPT   1
#define BR_QUEUE    2

#define DNS_PORT  53

#define OK              1
#define ERROR           0
#define ERROR_INDEX     -1

#define BOOL        int
#define BYTE        unsigned char
#define UINT16      unsigned short  /* 2-byte */
#define UINT32      unsigned int    /* 4-byte */

#define C_MAX_DNS_PACKET_SIZE   1500
#define C_MAX_URL_SIZE          63
#define DNS_PORT                53

#define DNS_QUERY       0
#define DNS_REPLY       1

#define DNS_QUERY_TYPE_IP       1
#define DNS_QUERY_TYPE_IPV6     28 /* Foxconn added by zacker, 04/23/2008,@ipv6_dns_redirect */
#define DNS_QUERY_TYPE_PTR      12
#define DNS_QUERY_TYPE_CNAME    5 

typedef struct S_IPHeader
{
    BYTE vers_len;
    BYTE type_of_services;
    BYTE length[2];
    BYTE ident[2];
    BYTE fragment[2];
    BYTE time;
    BYTE proto;
    BYTE checksum[2];
    BYTE src_ip[4];
    BYTE dst_ip[4];
    BYTE payload[1];
} T_IPHeader;

typedef struct S_UDPHeader
{
    BYTE src_port[2];
    BYTE dst_port[2];
    BYTE length[2];
    BYTE checksum[2];
    BYTE payload[1];
}T_UDPHeader;

#define UDP_PROTOCOL        0x11
#define IsUdpPacket(ptIp) (ptIp->proto == UDP_PROTOCOL)
#define GetIpHeaderLen(ptIp) (((ptIp->vers_len)&0x0f) << 2)
#define GetIpPayload(ptIp) ((BYTE *)ptIp + GetIpHeaderLen(ptIp))
#define GetIpPayloadLen(ptIp) (((ptIp->length[0]<<8) + ptIp->length[1]) - GetIpHeaderLen(ptIp))
#define GetUdpPayload(ptIp) (GetIpPayload(ptIp) + 8)

#define AG_UINT16_TO_BUFFER(buff, src) {UINT16 tmp; tmp=src; memcpy((char*)(buff),(char*)(&tmp),2);}
#define AG_BUFFER_TO_UINT16(dest, buff, convfun) {memcpy((char*)(&dest), (char*)(buff), 2); dest=convfun(dest);}

typedef struct S_dnsHeader
{
    unsigned short   id:16,      /* fields (Bytes 1-2) query identification number */

 //(_BYTE_ORDER == _BIG_ENDIAN) /* foxconn modified, wenchia, 2007/01/04, WGT624v4 is a big-endian platform. */
/*#error BIG_ENDIAN*/                     

    /* fields in third byte */
            qr:1,       /* response flag */
            opcode:4,   /* purpose of message */
            aa:1,       /* authoritive answer */
            tc:1,       /* truncated message */
            rd:1,       /* recursion desired */

    /* fields in fourth byte */
            ra:1,       /* recursion available */
            unused:2,   /* unused bits */
            bf:1,       /* broadcast flag */
            rcode:4;    /* response code */
//#elif (_BYTE_ORDER == _LITTLE_ENDIAN)
//#error LITTLE_ENDIAN   
#if 0                  
/*/*#else
    /* fields in third byte */
            rd:1,       /* recursion desired */
            tc:1,       /* truncated message */
            aa:1,       /* authoritive answer */
            opcode:4,   /* purpose of message */
            qr:1,       /* response flag */

    /* fields in fourth byte */
            rcode:4,    /* response code */
            bf:1,       /* broadcast flag */
            unused:2,   /* unused bits */
            ra:1;       /* recursion available */
#endif

    /* remaining bytes */
    unsigned short numQuestion;       /* number of question entries */
    unsigned short numAnswer;         /* number of answer entries */
    unsigned short numAuthority;      /* number of authority entries */
    unsigned short numAdditional;     /* number of resource entries */

} dnsHeader_T;

typedef struct S_dnsResourceRecord
{
    unsigned short type;
    unsigned short class;
    unsigned long  ttl;
    unsigned short len;
    unsigned char  data[100];
} dnsResourceRecord_T;

#define UDP_GET_DSTPORT(uh)    ntohs(*(UINT16 *)(uh->dst_port))

#define free kfree
#define malloc(a) kmalloc(a, GFP_ATOMIC)

#define printf(s...) printk(KERN_EMERG s)

#if 0
#define BR_DNS_HIJACK_DEBUG(fmt, args...)  printk(KERN_EMERG "br_dns_hijack DEBUG: " fmt, ##args)
#define BR_DNS_HIJACK_ERROR                printk(KERN_EMERG "br_dns_hijack ERROR: " fmt, ##args)
#else
#define BR_DNS_HIJACK_DEBUG(fmt, args...)
#define BR_DNS_HIJACK_ERROR
#endif

#endif 
